/*
 * $Id: qmail-qsmtp.c,v 1.13 2004/03/10 08:58:23 tarkhil Exp $
 *
 * SMTP forwarder
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <syslog.h>
#include <errno.h>
#include <sys/socket.h>
#include <unp.h>
#include <stdarg.h>

#define SERVADDR "127.0.0.1"
#define SERVPORT 10024

void write_log (char *message)
{
    syslog (LOG_INFO,"%s", message); 
}

void write_log2 (const char *msg1, const char *msg2)
{
    syslog (LOG_INFO,"%s: %s", msg1, msg2); 
}

void write_log_errno (const char *message)
{
    syslog (LOG_INFO,"%s: %s", message, strerror(errno)); 
}

void die_status (int code, char *message)
{
    syslog(LOG_ERR,"FATAL: %s", message); 
    exit (code);
}

void die_log(int code, char *msg1, char *msg2) {
    syslog (LOG_ERR,"FATAL: %s: %s", msg1, msg2); 
    exit(code);
}

void die_errno( int code, const char *message) {
    syslog(LOG_ERR, "FATAL: %s",message);
    exit(code);
}
char ** envptr;
const char * envnames[] = {
  "TCPREMOTEIP",
  "TCPREMOTEHOST",
  "RELAYCLIENT",
  NULL
};

void dumpenv() {
  const char **ep;
    for (ep = envnames; *ep; ep++) {
      syslog(LOG_INFO, "%s = %s", *ep, getenv((const char *)*ep));
    }
}

/* assume files are opened already */
int store(int from, int to) 
{
#define BUFSIZE 4096
  char buffer[BUFSIZE+1];
  int nread, nwrite = 0;
  const char ** ep;
  while ((nread = read(from, buffer, BUFSIZE)) != 0) {
    nwrite++;
    if (nread < 0) 
      die_errno(81, "couldn't read message");
    if (write(to, buffer, nread) < nread) 
      die_errno(81, "couldn't write temp message");
  }
  if (!nwrite) {
    /* Empty envelope? How can it be? */
    /* Probabily, spammer timed out? */
    syslog(LOG_INFO, "Empty envelope discovered");
    dumpenv();
  }
  return nwrite;
}

typedef struct {
    int code;
#define REPLY_LENGTH 128
    char reply[REPLY_LENGTH+1];
} smtp_reply;

/* kinf of dumb, but should work */
int get_reply(int fd, smtp_reply *reply) 
{
    int rc;
    syslog(LOG_DEBUG,"get_reply is waiting");
    if ((rc = readline(fd, reply->reply, REPLY_LENGTH)) <= 0) {
	return rc;
    }
    reply->code = atoi(reply->reply);
    syslog(LOG_DEBUG,"Got %d %s", reply->code, reply->reply);
    return rc;
}

int writeline(int fd, const char * line)
{
  /* allow empty mail from. Do we really need it? */
  if (strlen(line) == 0)  return 1;
    return(write(fd, line, strlen(line)));
}

void writeordie(int fd, const char * diag, const char *line, ...) {
    va_list ap;
    va_start(ap, line);
    while (line != NULL) {
      syslog(LOG_DEBUG, "writeordie: %s", line);
	if (writeline(fd, line) <= 0) {
	    die_errno(81, diag);
	}
	line = va_arg(ap, const char *);
    }
    syslog(LOG_DEBUG, "writeordie finished");
}

int smtpcopy(int from, int to) {
    char buffer[BUFSIZE+1];
    int nread, nwrite = 0;
    syslog(LOG_DEBUG, "Reading lines...");
#if 0
    while (!feof(stdin)) {
	if (fgets(buffer, BUFSIZE, stdin) == NULL) {
	  syslog(LOG_INFO,"EOF or error on stdin");
	    if (feof(stdin)) return nwrite;
#else
	while ((nread = readline(from, buffer, BUFSIZE)) != 0) {
	  if (nread < 0) {
#endif
	    die_errno(81, "Cannot read message body");
	  }
	syslog(LOG_DEBUG, "Writing %s", buffer);
	if (*buffer == '.') {
	    syslog(LOG_DEBUG, "escape");
	    if ((nwrite = write(to,".",1)) <= 0) {
		die_errno(81, "Cannot escape dot");
	    }
	}
	if ((nwrite = writeline(to, buffer)) < 0) {
	  syslog(LOG_INFO, "writeline '%s' returned %d", buffer, nwrite);
	    die_errno(81, "Cannot write message body");
	}
    }
    return nwrite;
}


int main(int argc, char ** argv, char **envp) 
{
    pid_t pid;
    char * envbuf = NULL;
    char * envelope;
    int envlen = 0;
    char buffer[BUFSIZE];
    int r_len;
    int efd;
    int sockfd;
    smtp_reply servreply;
    struct sockaddr_in server;
    const char * mailfrom = "MAIL FROM:<";
    const char * closeaddr = ">";
    const char * newline = "\r\n";
    const char * rcptto = "RCPT TO:<";
    const char * data = "DATA";
    const char * dot = ".";
    const char * quit = "QUIT";
    struct hostent *hp;
    int frombuf = 0;
    int attempt = 0;
    fd_set setdebug;
    struct timeval tmout;
    envptr = envp;
    openlog ("qmail-qsmtp", LOG_PID | LOG_NDELAY, LOG_MAIL); 
    dumpenv();
#ifdef TRY_ENVELOPE_FIRST
    FD_ZERO(&setdebug);
    FD_SET(0,&setdebug);
    FD_SET(1,&setdebug);
    tmout.tv_sec = 1;
    tmout.tv_usec = 0;
    syslog(LOG_INFO, "Selecting");
    switch (select(2,&setdebug, NULL, NULL, &tmout)) {
    case -1:
      die_errno(81, "Select failed");
    case 0:
      syslog(LOG_INFO, "Timeout in select...");
      break;
    default:
      if (FD_ISSET(1,&setdebug)) {
	syslog(LOG_INFO,"Descriptor 1 is ready");
	break;
      }
      if (FD_ISSET(0,&setdebug)) {
	syslog(LOG_INFO,"Descriptor 0 is ready");
#endif
	/* I'll have to buffer input file :-( */
	sprintf(buffer, "/tmp/tmp-queue.%ld", getpid());
	if ((frombuf = open(buffer, O_CREAT|O_RDWR|O_EXCL, 0600)) < 0) {
	  die_errno(81, "cannot create tempfile");
	}
	unlink(buffer);
	store(0, frombuf);
	close(0);
	if (lseek(frombuf, 0l, SEEK_SET) < 0) {
	  die_errno(81, "cannot rewind temp file");
	}
#ifdef TRY_ENVELOPE_FIRST
	syslog(LOG_INFO, "Data buffered, waiting for envelope");
      }
    }

    FD_ZERO(&setdebug);
    FD_SET(1,&setdebug);
    tmout.tv_sec = 1;
    tmout.tv_usec = 0;
    syslog(LOG_INFO, "Selecting again");
    switch (select(1,&setdebug, NULL, NULL, &tmout)) {
    case -1:
      die_errno(81, "Select failed");
    case 0:
      //      die_status(81, "Fatal timeout in select");
    default:
      /* happy */
    }
#endif
    while ((r_len = read(1, buffer, BUFSIZE)) != 0) {
      if ((envbuf = realloc(envbuf, envlen+r_len)) == NULL) {
	die_errno(51, "Not enough memory for environment");
      }
      memcpy(envbuf+envlen, buffer, r_len);
      envlen += r_len;
    }
    /* Quite simple; just parse environment and push everything else by SMTP */
    /* First of all, connect to (dumb) localhost:10024 */
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	die_errno(81, "couldn't create socket");
    }
    bzero(&server, sizeof(server));
    server.sin_family = AF_INET;
    server.sin_port = htons(SERVPORT);
    if (inet_pton(AF_INET, SERVADDR, &server.sin_addr) <= 0) {
	die_errno(81, "inet_pton failed");
    }
    if (connect(sockfd, (SA *)&server, sizeof(server)) < 0) {
	die_errno(81, "connect failed");
    }
    syslog(LOG_DEBUG, "Connected to SMTP");
    /* 250 ESMTP */
    if (get_reply(sockfd, &servreply) <= 0) {
	die_errno(81, "No SMTP greeting");
    }
    if (servreply.code != 220) {
	die_log(81, "bad greeting", servreply.reply);
    }
    /* I'll implement HELO later */
    envelope = envbuf;
    if (*envelope != 'F') {
	die_errno(91, "F expected");
    }
    envelope++;
    syslog(LOG_INFO, "writing 'mail from' %s", envelope);
    writeordie(sockfd,"writing mail from", mailfrom, envelope, closeaddr, newline, NULL);
    if (get_reply(sockfd, &servreply) <= 0) {
	die_errno(81, "No SMTP reply");
    }
    if (servreply.code != 250) {
	die_log(81, "bad mailfrom", servreply.reply);
    }
    envelope += strlen(envelope)+1;
    while (*envelope) {
	if (*envelope != 'T') {
	    die_errno(91, "T expected");
	}
	envelope++;
	writeordie(sockfd,"writing rcpt to", rcptto, envelope, closeaddr, newline, NULL);
	syslog(LOG_INFO, "rcpt to %s", envelope); 
	if (get_reply(sockfd, &servreply) <= 0) {
	    die_errno(81, "No SMTP reply");
	}
	if (servreply.code != 250) {
	    die_log(81, "bad rcptto", servreply.reply);
	}
	envelope += strlen(envelope)+1;
    }
    /* Okay, I want to send data now... */
    writeordie(sockfd, "asking for data", data, newline, NULL);
    if (get_reply(sockfd, &servreply) <= 0) {
	die_errno(81, "No SMTP reply");
    }
    if (servreply.code != 354) {
	die_log(81, "they don't want to listen", servreply.reply);
    }
    /* Dammit. I need to escape single dots */
    syslog(LOG_DEBUG, "Pushing body");
    smtpcopy(frombuf, sockfd);
    if (frombuf) close(frombuf);
    syslog(LOG_DEBUG, "Okay, going to finish");
    writeordie(sockfd, "finishing data", newline, dot, newline, NULL);
    if (get_reply(sockfd, &servreply) <= 0) {
	die_errno(81, "No SMTP reply");
    }
    writeordie(sockfd, "quitting", quit, newline, NULL);
    close(sockfd);
    if (servreply.code >= 500) {
	syslog (LOG_NOTICE,"amavis permanently refused to send message from %s (%s): %s", 
		envbuf+1, getenv("TCPREMOTEIP"),servreply.reply); 
	exit(31);
    }
    if (servreply.code >= 400) {
	syslog (LOG_NOTICE,"amavis temporarily refused to send message from %s (%s): %s", 
		envbuf+1, getenv("TCPREMOTEIP"), servreply.reply); 
	exit(71);
    }
    /* store message */
    syslog (LOG_NOTICE,"amavis passed message from %s (%s): %s", 
	    envbuf+1, getenv("TCPREMOTEIP"), servreply.reply); 
    free(envbuf);
    closelog (); 
    exit(0);
}
